<?php

namespace Drupal\backfill_formatter;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Defines a class for backfilling entity-reference values using common terms.
 */
class BackFillTerms {

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * Database.
   *
   * @var \Drupal\Core\Database\Connection
   */
  private $database;

  /**
   * Current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  private $currentUser;

  /**
   * Query plugin manager.
   *
   * @var \Drupal\backfill_formatter\BackFillQueryPluginManager
   */
  private $backFillQueryPluginManager;

  /**
   * Constructs a new BackfillTerms.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity type manager.
   * @param \Drupal\Core\Database\Connection $database
   *   Database.
   * @param \Drupal\Core\Session\AccountInterface $currentUser
   *   Current user.
   * @param \Drupal\backfill_formatter\BackFillQueryPluginManager $backFillQueryPluginManager
   *   Query plugin manager.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, Connection $database, AccountInterface $currentUser, BackFillQueryPluginManager $backFillQueryPluginManager) {
    $this->entityTypeManager = $entityTypeManager;
    $this->database = $database;
    $this->currentUser = $currentUser;
    $this->backFillQueryPluginManager = $backFillQueryPluginManager;
  }

  /**
   * Backfills entities.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Entity to backfill.
   * @param string $field_name
   *   Field to backfill.
   * @param int $backfill_count
   *   Backfill count.
   * @param array $vocabulary_weights
   *   Array of weights, keyed by vocabulary ID.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   Array of entities.
   */
  public function backfillEntity(ContentEntityInterface $entity, $field_name, int $backfill_count, array $vocabulary_weights = []) : array {
    if (!$entity->hasField($field_name)) {
      return [];
    }
    $field_definition = $entity->getFieldDefinition($field_name);
    $target_entity_type_id = $field_definition->getSetting('target_type');
    $values = array_column($entity->get($field_name)->getValue(), 'target_id');
    $entity_storage = $this->entityTypeManager->getStorage($target_entity_type_id);
    if (count($values) >= $backfill_count) {
      return $entity_storage->loadMultiple($values);
    }
    $query = $this->database->select('taxonomy_entity_index', 'tei')->fields('tei', ['tid'])
      ->condition('entity_type', $entity->getEntityTypeId())
      ->condition('entity_id', $entity->id());
    if ($revision_id = $entity->getRevisionId()) {
      $query->condition('revision_id', $revision_id);
    }
    $entity_term_ids = $query->execute()->fetchCol();
    if (empty($entity_term_ids)) {
      // Entity has no terms.
      return $entity_storage->loadMultiple($values);
    }
    $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
    $target_bundles = $field_definition->getSetting('handler_settings')['target_bundles'] ?? [];
    $remaining = $backfill_count - count($values);
    if (empty($vocabulary_weights)) {
      $values = array_merge($values, $this->queryMatchingEntities($entity, $target_entity_type_id, $entity_term_ids, $target_bundles, $remaining, $values));
      return $entity_storage->loadMultiple($values);
    }
    $tid_vid_pairs = $term_storage->getAggregateQuery()
      ->condition('tid', $entity_term_ids, 'IN')
      ->groupBy('vid')
      ->groupBy('tid')
      ->execute();
    asort($vocabulary_weights);
    while ($vocabulary_weights && $remaining) {
      $remaining = $backfill_count - count($values);
      // The vocabulary IDs are the keys.
      reset($vocabulary_weights);
      $vocabulary = key($vocabulary_weights);
      array_shift($vocabulary_weights);
      $vocabulary_term_ids = array_map(function (array $item) {
        return $item['tid'];
      }, array_filter($tid_vid_pairs, function (array $item) use ($vocabulary) {
        return $item['vid'] === $vocabulary;
      }));
      if (empty($vocabulary_term_ids)) {
        continue;
      }
      $values = array_merge($values, $this->queryMatchingEntities($entity, $target_entity_type_id, $vocabulary_term_ids, $target_bundles, $remaining, $values));
    }
    return $entity_storage->loadMultiple($values);
  }

  /**
   * Query for matching entities.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Host entity being back-filled.
   * @param string $target_entity_type_id
   *   Type of entity being reference.
   * @param array $common_term_ids
   *   Term IDs to match on.
   * @param array $target_bundles
   *   Target bundles to match on.
   * @param int $length
   *   Number of matches required.
   * @param array $exclude
   *   Array of entities to exclude.
   *
   * @return array
   *   Array of entity IDs.
   */
  private function queryMatchingEntities(ContentEntityInterface $entity, string $target_entity_type_id, array $common_term_ids, array $target_bundles, int $length, array $exclude) : array {
    $exclude[] = $entity->id();
    $query = $this->database->select('taxonomy_entity_index', 'tei')
      ->fields('tei', ['entity_id'])
      ->condition('entity_type', $target_entity_type_id)
      ->condition('entity_id', $exclude, 'NOT IN')
      ->condition('tid', $common_term_ids, 'IN')
      ->groupBy('entity_id');
    $alias = $query->addExpression('COUNT(entity_id)', 'matches');
    // Order by most matches, then tid and entity id (for consistent results).
    $query->orderBy($alias, 'DESC')
      ->orderBy('tid', 'ASC')
      ->orderBy('entity_id', 'DESC');
    if ($target_bundles) {
      $query->condition('bundle', $target_bundles, 'IN');
    }
    $query->range(0, $length);
    try {
      /** @var \Drupal\backfill_formatter\BackFillQueryPluginInterface $alter */
      $alter = $this->backFillQueryPluginManager->createInstance('default:' . $target_entity_type_id);
      $alter->alterQuery($this->entityTypeManager, $query, $this->currentUser, $entity->language()->getId());
    }
    catch (PluginException $e) {
      // Continue.
    }
    return $query->execute()->fetchCol();
  }

}
